﻿using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Threading;

namespace TestProject
{
    /// <summary>
    /// 비동기 가상화 컬렉션
    /// </summary>
    /// <typeparam name="TItem">항목 타입</typeparam>
    public class AsyncVirtualizingCollection<TItem> : VirtualizingCollection<TItem>, INotifyCollectionChanged, INotifyPropertyChanged
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 속성 변경시 이벤트 - PropertyChanged

        /// <summary>
        /// 속성 변경시 이벤트
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion
        #region 컬렉션 변경시 이벤트 - CollectionChanged

        /// <summary>
        /// 컬렉션 변경시 이벤트
        /// </summary>
        public event NotifyCollectionChangedEventHandler CollectionChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 동기화 컨텍스트
        /// </summary>
        private readonly SynchronizationContext synchronizationContext;

        /// <summary>
        /// 로딩 여부
        /// </summary>
        private bool isLoading;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 동기화 컨텍스트 - SynchronizationContext

        /// <summary>
        /// 동기화 컨텍스트
        /// </summary>
        protected SynchronizationContext SynchronizationContext
        {
            get
            {
                return this.synchronizationContext;
            }
        }

        #endregion
        #region 로딩 여부 - IsLoading

        /// <summary>
        /// 로딩 여부
        /// </summary>
        public bool IsLoading
        {
            get
            {
                return this.isLoading;
            }
            set
            {
                if(value != this.isLoading)
                {
                    this.isLoading = value;
                }

                FirePropertyChangedEvent("IsLoading");
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - AsyncVirtualizingCollection(itemProvider)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="itemProvider">항목 공급자</param>
        public AsyncVirtualizingCollection(IItemProvider<TItem> itemProvider) : base(itemProvider)
        {
            this.synchronizationContext = SynchronizationContext.Current;
        }

        #endregion
        #region 생성자 - AsyncVirtualizingCollection(itemProvider, pageSize)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="itemProvider">항목 공급자</param>
        /// <param name="pageSize">페이지 크기</param>
        public AsyncVirtualizingCollection(IItemProvider<TItem> itemProvider, int pageSize) : base(itemProvider, pageSize)
        {
            this.synchronizationContext = SynchronizationContext.Current;
        }

        #endregion
        #region 생성자 - AsyncVirtualizingCollection(itemProvider, pageSize, pageTimeout)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="itemProvider">항목 공급자</param>
        /// <param name="pageSize">페이지 크기</param>
        /// <param name="pageTimeout">페이지 타임아웃</param>
        public AsyncVirtualizingCollection(IItemProvider<TItem> itemProvider, int pageSize, int pageTimeout) : base(itemProvider, pageSize, pageTimeout)
        {
            this.synchronizationContext = SynchronizationContext.Current;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 속성 변경시 이벤트 발생시키기 - FirePropertyChangedEvent(e)

        /// <summary>
        /// 속성 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected virtual void FirePropertyChangedEvent(PropertyChangedEventArgs e)
        {
            PropertyChanged?.Invoke(this, e);
        }

        #endregion
        #region 컬렉션 변경시 이벤트 발생시키기 - FireCollectionChangedEvent(e)

        /// <summary>
        /// 컬렉션 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        protected virtual void FireCollectionChangedEvent(NotifyCollectionChangedEventArgs e)
        {
            CollectionChanged?.Invoke(this, e);
        }

        #endregion

        #region 카운트 로드하기 - LoadCount()

        /// <summary>
        /// 카운트 로드하기
        /// </summary>
        protected override void LoadCount()
        {
            Count = 0;

            IsLoading = true;

            ThreadPool.QueueUserWorkItem(LoadCount);
        }

        #endregion
        #region 페이지 로드하기 - LoadPage(index)

        /// <summary>
        /// 페이지 로드하기
        /// </summary>
        /// <param name="index">인덱스</param>
        protected override void LoadPage(int index)
        {
            IsLoading = true;

            ThreadPool.QueueUserWorkItem(LoadPage, index);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 속성 변경시 이벤트 발생시키기 - FirePropertyChangedEvent(propertyName)

        /// <summary>
        /// 속성 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="propertyName">속성명</param>
        private void FirePropertyChangedEvent(string propertyName)
        {
            FirePropertyChangedEvent(new PropertyChangedEventArgs(propertyName));
        }

        #endregion
        #region 컬렉션 변경시 리셋 이벤트 발생시키기 - FireCollectionChangedResetEvent()

        /// <summary>
        /// 컬렉션 변경시 리셋 이벤트 발생시키기
        /// </summary>
        private void FireCollectionChangedResetEvent()
        {
            FireCollectionChangedEvent(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
        }

        #endregion

        #region 카운트 로드하기 - LoadCount(argument)

        /// <summary>
        /// 카운트 로드하기
        /// </summary>
        /// <param name="argument">인자</param>
        private void LoadCount(object argument)
        {
            int count = FetchCount();

            SynchronizationContext.Send(LoadCountCompleted, count);
        }

        #endregion
        #region 카운트 로드 완료시 처리하기 - LoadCountCompleted(argument)

        /// <summary>
        /// 카운트 로드 완료시 처리하기
        /// </summary>
        /// <param name="argument">인자</param>
        private void LoadCountCompleted(object argument)
        {
            Count = (int)argument;

            IsLoading = false;

            FireCollectionChangedResetEvent();
        }

        #endregion
        #region 페이지 로드하기 - LoadPage(argument)

        /// <summary>
        /// 페이지 로드하기
        /// </summary>
        /// <param name="argument">인자</param>
        private void LoadPage(object argument)
        {
            int pageIndex = (int)argument;

            IList<TItem> itemList = FetchPage(pageIndex);

            SynchronizationContext.Send(LoadPageCompleted, new object[]{ pageIndex, itemList });
        }

        #endregion
        #region 페이지 로드 완료시 처리하기 - LoadPageCompleted(argument)

        /// <summary>
        /// 페이지 로드 완료시 처리하기
        /// </summary>
        /// <param name="argument">인자</param>
        private void LoadPageCompleted(object argument)
        {
            int pageIndex = (int)((object[]) argument)[0];

            IList<TItem> itemList = (IList<TItem>)((object[])argument)[1];

            PopulatePage(pageIndex, itemList);

            IsLoading = false;

            FireCollectionChangedResetEvent();
        }

        #endregion
    }
}